package main

import (
	"bufio"
	"fmt"
	"os"
	"path/filepath"
	"strings"
	"time"

	"github.com/tidwall/sjson"

	"github.com/vmware-labs/yaml-jsonpath/pkg/yamlpath"

	"github.com/Masterminds/semver"
	"github.com/go-git/go-git/v5"
	"github.com/tidwall/gjson"

	"gopkg.in/yaml.v3"
)

var (
	content     []byte
	currentType string
	currentYaml yaml.Node
	permissions os.FileMode
)

func readFileAndInitPermissions() error {
	fileStats, err := os.Stat(arguments.Path)
	if err != nil {
		return err
	}

	permissions = fileStats.Mode().Perm()

	content, err = os.ReadFile(arguments.Path)
	return err
}

func getFromJson() (string, error) {

	err := readFileAndInitPermissions()
	if err != nil {
		return "", err
	}

	value := gjson.Get(string(content), arguments.Selector)

	if value.Type == gjson.Null {
		return "", notFoundError
	}

	if t := value.Type; t != gjson.String {
		return "", noSemverError
	}

	return value.String(), nil
}

func getFromYaml() (string, error) {

	err := readFileAndInitPermissions()
	if err != nil {
		return "", err
	}

	var node yaml.Node
	if err := yaml.Unmarshal(content, &node); err != nil {
		return "", err
	}

	currentYaml = node

	path, err := yamlpath.NewPath(arguments.Selector)
	if err != nil {
		return "", err
	}

	results, err := path.Find(&node)
	if err != nil {
		return "", err
	}

	// len() for []*gopkg.in/yaml.v3.Node is defined as zero (S1009)
	if len(results) == 0 {
		return "", notFoundError
	}

	if len(results) > 1 {
		return "", multipleFoundError
	}

	result := results[0]
	return result.Value, nil

}

func readVersion() (string, error) {

	if arguments.Git {
		version, err := getLatestSemanticTag()
		return version.String(), err

	} else if arguments.Path != "" {

		if arguments.Selector == "" {
			return "", missingSelector
		}

		ext := filepath.Ext(arguments.Path)
		switch ext {
		case ".json":
			currentType = "json"
			return getFromJson()
		case ".yaml":
			currentType = "yaml"
			return getFromYaml()
		default:
			return "", notImplementedError
		}

	}

	return readVersionFromStdIn()

}

func readVersionFromStdIn() (string, error) {

	version := ""

	inChannel := make(chan string)
	errChannel := make(chan error, 1)

	defer close(inChannel)
	defer close(errChannel)

	go func(c chan string, e chan error) {

		reader := bufio.NewReader(os.Stdin)
		for {
			s, err := reader.ReadBytes('\n')
			if err != nil {
				e <- err
				return
			}
			c <- string(s)
		}

	}(inChannel, errChannel)

loop:
	for {
		select {
		case error := <-errChannel:
			fmt.Println("error:", error)
			return version, error

		case v := <-inChannel:
			version += v
			break loop

		case <-time.After(1 * time.Minute):
			return version, timeoutError
		}

	}

	return version, nil
}

func getVersion() (*semver.Version, error) {

	v, err := readVersion()

	if err != nil {
		return &semver.Version{}, err
	}

	v = strings.TrimSpace(v)
	return semver.NewVersion(v)

}

func writeVersionToJson(version string) error {
	if content == nil {
		return fmt.Errorf("content is nil")
	}

	result, err := sjson.Set(string(content), arguments.Selector, version)
	if err != nil {
		return err
	}

	err = os.WriteFile(arguments.Path, []byte(result), permissions)

	return err
}

func writeVersionToYaml(version string) error {

	path, err := yamlpath.NewPath(arguments.Selector)
	if err != nil {
		return err
	}
	results, err := path.Find(&currentYaml)
	if err != nil {
		return err
	}

	results[0].Value = version

	//results.Value = version
	//fmt.Println(results)

	result, err := yaml.Marshal(&currentYaml)
	if err != nil {
		return err
	}

	err = os.WriteFile(arguments.Path, []byte(result), permissions)

	return err
}

func writeVersionToGit(version string) error {

	repo := gitRepo

	h, err := repo.Head()
	if err != nil {
		return err
	}

	if verbosity {
		fmt.Println("Writing to git HEAD:", h.Hash())
	}

	_, err = repo.CreateTag(version, h.Hash(), &git.CreateTagOptions{
		Message: "chore: bump version to " + version,
	})

	if err != nil {
		return err
	}

	return nil
}

func writeVersion(version string) error {

	if predict {
		if nullTerminated {
			version += string(rune(0))
			fmt.Printf("%s", version)
		} else {
			fmt.Printf(version)
		}
		return nil
	}

	if arguments.Git {
		return writeVersionToGit(version)
	} else if currentType == "json" {
		return writeVersionToJson(version)
	} else if currentType == "yaml" {
		return writeVersionToYaml(version)
	}

	if arguments.Path == "" {
		if nullTerminated {
			version += string(rune(0))
			fmt.Printf("%s", version)
		} else {
			fmt.Printf(version)
		}
		return nil
	}

	msg := fmt.Sprintf("Unknown type: %s for path: %s", currentType, arguments.Path)
	_, err := fmt.Fprintf(os.Stderr, "%s\n", msg)
	if err != nil {
		fmt.Printf("%s\n", msg)
	}

	return nil

}
